//
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#define LINEWIDTH   60
#define INDENT_SIZE 2
#define INDENT_STRING " "

// Compute stack of Crc's for each subdirectory encountered.
unsigned long CrcStack[256];


/* crc32h.c -- package to compute 32-bit CRC one byte at a time using   */
/*             the high-bit first (Big-Endian) bit ordering convention  */
/*                                                                      */
/* Synopsis:                                                            */
/*  gen_crc_table() -- generates a 256-word table containing all CRC    */
/*                     remainders for every possible 8-bit byte.  It    */
/*                     must be executed (once) before any CRC updates.  */
/*                                                                      */
/*  unsigned update_crc(crc_accum, data_blk_ptr, data_blk_size)         */
/*           unsigned crc_accum; char *data_blk_ptr; int data_blk_size; */
/*           Returns the updated value of the CRC accumulator after     */
/*           processing each byte in the addressed block of data.       */
/*                                                                      */
/*  It is assumed that an unsigned long is at least 32 bits wide and    */
/*  that the predefined type char occupies one 8-bit byte of storage.   */
/*                                                                      */
/*  The generator polynomial used for this version of the package is    */
/*  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x^1+x^0 */
/*  as specified in the Autodin/Ethernet/ADCCP protocol standards.      */
/*  Other degree 32 polynomials may be substituted by re-defining the   */
/*  symbol POLYNOMIAL below.  Lower degree polynomials must first be    */
/*  multiplied by an appropriate power of x.  The representation used   */
/*  is that the coefficient of x^0 is stored in the LSB of the 32-bit   */
/*  word and the coefficient of x^31 is stored in the most significant  */
/*  bit.  The CRC is to be appended to the data most significant byte   */
/*  first.  For those protocols in which bytes are transmitted MSB      */
/*  first and in the same order as they are encountered in the block    */
/*  this convention results in the CRC remainder being transmitted with */
/*  the coefficient of x^31 first and with that of x^0 last (just as    */
/*  would be done by a hardware shift register mechanization).          */
/*                                                                      */
/*  The table lookup technique was adapted from the algorithm described */
/*  by Avram Perez, Byte-wise CRC Calculations, IEEE Micro 3, 40 (1983).*/

#define POLYNOMIAL 0x04c11db7L

static unsigned long crc_table[256];

/* generate the table of CRC remainders for all possible bytes */
void gen_crc_table(void)
{
  register int i, j;
  register unsigned long crc_accum;

  for ( i = 0;  i < 256;  i++ ) {
    crc_accum = ( (unsigned long) i << 24 );

    for ( j = 0;  j < 8;  j++ ) {
      if ( crc_accum & 0x80000000L )
        crc_accum = ( crc_accum << 1 ) ^ POLYNOMIAL;
      else
        crc_accum = ( crc_accum << 1 );
      }

    crc_table[i] = crc_accum;
    }
}

/* update the CRC on the data block one byte at a time */
unsigned long compute_crc(unsigned long crc_accum, char *data_blk_ptr, int data_blk_size)
{
  register int i, j;

  for ( j = 0;  j < data_blk_size;  j++ ) {
    i = ( (int) ( crc_accum >> 24) ^ *data_blk_ptr++ ) & 0xff;
    crc_accum = ( crc_accum << 8 ) ^ crc_table[i];
    }

  return crc_accum;
}

void ProcessFile(char *filename, int indent)
{
	FILE *f;
	char buffer[8192];
	unsigned long bytes;

	// Reset the CRC for this level; all subdirectories above will
	// continue to accumulate for this file, though.
	CrcStack[indent] = 0;

	f = fopen(filename,"rb");
	if (!f) 
	{
		printf("%*s : *** Cannot Open\n", -LINEWIDTH, filename);
		return;
	}

	do 
	{
		bytes = fread(buffer,1,sizeof(buffer),f);
		if (bytes) 
		{
			for (int i=0; i<=indent; i++)
			{
				CrcStack[i] = compute_crc(CrcStack[i], buffer, bytes);
			}
		}
	} while (bytes);

	fclose(f);
	printf("%*s : %08lX\n", -LINEWIDTH, filename, CrcStack[indent]);
}

void ProcessName(char *filename, int indent)
{
	char			current_name[MAX_PATH];
	HANDLE			hFind;
	WIN32_FIND_DATA finddata;

	// Reset the CRC for this level; all subdirectories above will
	// continue to accumulate for this file, though.
	CrcStack[indent] = 0;

	// Try to determine if this is a directory first.
	sprintf(current_name,"%s\\*.*", filename);
	hFind = FindFirstFile(current_name, &finddata);

	// If it isn't a directory, we will have failed the above; try the plain file next.
	if (hFind==INVALID_HANDLE_VALUE)
	{
		hFind = FindFirstFile(filename,&finddata);
	}

	// If both of those failed, we really can't open the file; quit this routine.
	if (hFind==INVALID_HANDLE_VALUE)
	{
		printf("%*s : *** Cannot Open\n", -LINEWIDTH, filename);
		return;
	}

	// Else, process the file(s)
	do
	{
		// Create a new "decended" copy with a separating backslash.
		sprintf(current_name,"%s\\%s", filename, finddata.cFileName);

		if (finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			if (strcmp(finddata.cFileName,".")!=0 &&
				strcmp(finddata.cFileName,"..")!=0)
			{
				ProcessName(current_name, indent+1);
			}
		}
		else
		{
			ProcessFile(current_name, indent+1);
		}
	}
	while (FindNextFile(hFind, &finddata));

	FindClose(hFind);
	printf("%*s =>%08lX\n", -LINEWIDTH, filename, CrcStack[indent]);
}

void main(int argc, char *argv[])
{
	if (argc<2) 
	{
		printf("Usage: crc32 <filename(s) or directory(s)>\n");
		printf("Prints 32-bit CRCs used to comparing files without the actual compare operation\n");
		exit(1);
	}

	gen_crc_table();

	for (int i=1; i<argc; i++)
	{
		ProcessName(argv[i], 0);
	}
}
